import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.nio.charset.Charset;
import java.util.ArrayList;

import javax.sound.sampled.AudioInputStream;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.Clip;
import javax.sound.sampled.LineUnavailableException;
import javax.sound.sampled.Mixer;
import javax.sound.sampled.UnsupportedAudioFileException;

public class Audio extends Thread {
    public static int SAMPLING_RATE = 48000;
    public static int SOUND_FREQ = 880;
    private int times;
    private int msec;
    private Mixer.Info mixerInfo;
    private AudioInputStream audio;
    
    public Audio(int times, int msec, String mixerName) {
        this.times = times;
        this.msec  = msec;
        audio = new WaveBeep(SAMPLING_RATE, SOUND_FREQ, msec/2);
        mixerInfo = getMixerByName(mixerName);
    }
    
    public Audio(String file, String mixerName) {
    	try {
    		audio = AudioSystem.getAudioInputStream(new File(file));
		} catch (UnsupportedAudioFileException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
        mixerInfo = getMixerByName(mixerName);
    }
    
    public Mixer.Info getMixerByName(String name) {
        Mixer.Info[] info = AudioSystem.getMixerInfo();

    	for (int i=0; i<info.length; i++) {
    		if (deMessyCode(info[i].getName()).equals(name)) return info[i];
    	}
    	return null;
    }

    public static String[] getMixerNames() {
        System.out.println("Mixer List");
        Mixer.Info[] info = AudioSystem.getMixerInfo();
    	String[] ret = new String[info.length];
    	for (int i=0; i<info.length; i++) {
    		ret[i] = deMessyCode(info[i].getName());
    	}
    	return ret;
    }
    
    @Override
    public void run() {
    	//System.out.println("Sound Start");
    	System.out.println("Sound " + times + " * " + msec + "ms");
        try {
            Clip clip = AudioSystem.getClip(mixerInfo);
            clip.open(audio);
            if (times > 0) {
            	for (int i=0; i<times; i++) {
            		clip.setFramePosition(0);
            		clip.start();
            		sleep(msec);
            	}
            } else {
        		clip.setFramePosition(0);
        		clip.start();
            }
            clip.close();
        } catch (LineUnavailableException e) {
            e.printStackTrace();
        } catch (IOException e) {
            e.printStackTrace();
        } catch (InterruptedException e) {
            e.printStackTrace();
        } catch (NegativeArraySizeException e) {
        	//e.printStackTrace(); // PulseAudioClip.java:450
        }
        //System.out.println("Sound End");
    }
    
    /*
     * JDKのバグ対策（サウンドデバイス名のエンコード間違いを直す）
     * https://qiita.com/TakaW/items/023b06cd1ba18ca62047
     * https://bugs.java.com/bugdatabase/view_bug.do?bug_id=JDK-8177951 
     */
    private static final byte HIGH_BIT = (byte) 0b11000000;
    private static String deMessyCode(String messyCode) {
        ByteArrayOutputStream buf = new ByteArrayOutputStream(messyCode.length());
        byte[] originalBytes = messyCode.getBytes(Charset.forName("UTF-8"));
        String systemEnc = System.getProperty("file.encoding");
        //System.out.println(systemEnc + (systemEnc.equalsIgnoreCase("MS932")));

        for (int i = 0; i < originalBytes.length; i++) {
            if ((byte) (originalBytes[i] & HIGH_BIT) == HIGH_BIT) {
            	if (i==originalBytes.length-2 && systemEnc.equalsIgnoreCase("MS932")) break;
            	// SJIS の泣き別れが起きないよう、最後の1文字は書き込まない
                buf.write((originalBytes[i] << 6) | (originalBytes[i + 1] << 2 >>> 2));
                // DELETE 0b110000 and move 6, then delete the 0b10 prefix of the second byte.
                i++;
            } else {
                buf.write(originalBytes[i]);
            }
        }
        return new String(buf.toByteArray(), Charset.forName(systemEnc));
    }
}
